import json

import pytest

from xcpcio.types import Team, Teams


class TestTeam:
    """Test cases for the Team dataclass"""

    def test_team_creation(self):
        """Test basic team creation with default values"""
        team = Team()
        assert team.id == ""
        assert team.name == ""
        assert team.organization is None
        assert team.members is None
        assert team.coaches is None
        assert team.location is None
        assert team.group == []
        assert team.extra == {}

    def test_team_creation_with_values(self):
        """Test team creation with provided values"""
        team = Team(
            id="team001",
            name="Test Team",
            organization="Test University",
            members=["Alice", "Bob", "Charlie"],
            coaches="Dr. Smith",
            location="Building A",
        )

        assert team.id == "team001"
        assert team.name == "Test Team"
        assert team.organization == "Test University"
        assert team.members == ["Alice", "Bob", "Charlie"]
        assert team.coaches == "Dr. Smith"
        assert team.location == "Building A"

    def test_group_management(self):
        """Test add_group and remove_group methods"""
        team = Team(id="test", name="Test Team", organization="Test Org")

        # Test adding groups
        team.add_group("undergraduate")
        assert "undergraduate" in team.group

        team.add_group("local")
        assert "local" in team.group
        assert len(team.group) == 2

        # Test that duplicate groups are not added
        team.add_group("undergraduate")
        assert len(team.group) == 2

        # Test removing groups
        team.remove_group("local")
        assert "local" not in team.group
        assert len(team.group) == 1

        # Test removing non-existent group (should not raise error)
        team.remove_group("nonexistent")
        assert len(team.group) == 1


class TestTeamSerialization:
    """Test cases for Team serialization and deserialization"""

    @pytest.fixture
    def sample_team(self) -> Team:
        """Create a sample team for testing"""
        return Team(
            id="team001",
            name="Alpha Team",
            organization="University A",
            members=["Alice", "Bob", "Charlie"],
            coaches="Dr. Smith",
            location="Building A",
            group=["undergraduate", "local"],
            extra={"room": "101", "contact": "alice@test.edu"},
        )

    def test_model_dump(self, sample_team: Team):
        """Test Team model_dump method"""
        team_dict = sample_team.model_dump()

        assert isinstance(team_dict, dict)
        assert team_dict["id"] == "team001"
        assert team_dict["name"] == "Alpha Team"
        assert team_dict["organization"] == "University A"
        assert team_dict["members"] == ["Alice", "Bob", "Charlie"]
        assert team_dict["coaches"] == "Dr. Smith"
        assert team_dict["location"] == "Building A"
        assert team_dict["group"] == ["undergraduate", "local"]
        assert "extra" not in team_dict  # extra field is excluded from serialization

    def test_model_validate(self, sample_team: Team):
        """Test Team model_validate method"""
        team_dict = sample_team.model_dump()
        reconstructed_team = Team.model_validate(team_dict)

        assert reconstructed_team.id == sample_team.id
        assert reconstructed_team.name == sample_team.name
        assert reconstructed_team.organization == sample_team.organization
        assert reconstructed_team.members == sample_team.members
        assert reconstructed_team.coaches == sample_team.coaches
        assert reconstructed_team.location == sample_team.location
        assert reconstructed_team.group == sample_team.group
        assert reconstructed_team.extra == {}  # extra is not serialized

    def test_model_dump_json(self, sample_team: Team):
        """Test Team model_dump_json method"""
        team_json = sample_team.model_dump_json()

        assert isinstance(team_json, str)
        # Verify it's valid JSON
        parsed = json.loads(team_json)
        assert parsed["id"] == "team001"
        assert parsed["name"] == "Alpha Team"

    def test_model_validate_json(self, sample_team: Team):
        """Test Team model_validate_json method"""
        team_json = sample_team.model_dump_json()
        reconstructed_team = Team.model_validate_json(team_json)

        assert reconstructed_team.id == sample_team.id
        assert reconstructed_team.name == sample_team.name
        assert reconstructed_team.organization == sample_team.organization
        assert reconstructed_team.members == sample_team.members
        assert reconstructed_team.coaches == sample_team.coaches
        assert reconstructed_team.location == sample_team.location
        assert reconstructed_team.group == sample_team.group
        assert reconstructed_team.extra == {}  # extra is not serialized

    def test_round_trip_dict(self, sample_team: Team):
        """Test complete round-trip through dict serialization"""
        team_dict = sample_team.model_dump()
        reconstructed_team = Team.model_validate(team_dict)

        assert reconstructed_team.id == sample_team.id
        assert reconstructed_team.name == sample_team.name
        assert reconstructed_team.organization == sample_team.organization
        assert reconstructed_team.members == sample_team.members
        assert reconstructed_team.coaches == sample_team.coaches
        assert reconstructed_team.location == sample_team.location
        assert reconstructed_team.group == sample_team.group
        assert reconstructed_team.extra == {}  # extra is excluded and won't round-trip

    def test_round_trip_json(self, sample_team: Team):
        """Test complete round-trip through JSON serialization"""
        team_json = sample_team.model_dump_json()
        reconstructed_team = Team.model_validate_json(team_json)

        assert reconstructed_team.id == sample_team.id
        assert reconstructed_team.name == sample_team.name
        assert reconstructed_team.organization == sample_team.organization
        assert reconstructed_team.members == sample_team.members
        assert reconstructed_team.coaches == sample_team.coaches
        assert reconstructed_team.location == sample_team.location
        assert reconstructed_team.group == sample_team.group
        assert reconstructed_team.extra == {}  # extra is excluded and won't round-trip

    def test_minimal_team_serialization(self):
        """Test serialization of team with default/minimal values"""
        minimal_team = Team()

        # Test dict round-trip
        team_dict = minimal_team.model_dump()
        reconstructed_from_dict = Team.model_validate(team_dict)
        assert reconstructed_from_dict == minimal_team

        # Test JSON round-trip
        team_json = minimal_team.model_dump_json()
        reconstructed_from_json = Team.model_validate_json(team_json)
        assert reconstructed_from_json == minimal_team


class TestTeamList:
    """Test cases for Teams (RootModel) serialization and deserialization"""

    @pytest.fixture
    def sample_teams(self) -> Teams:
        """Create a Teams instance for testing"""
        return Teams(
            [
                Team(
                    id="team001",
                    name="Alpha Team",
                    organization="University A",
                    members=["Alice", "Bob"],
                    coaches="Coach A",
                ),
                Team(
                    id="team002",
                    name="Beta Team",
                    organization="University B",
                    members=["Charlie", "David", "Eve"],
                    location="Remote",
                ),
                Team(
                    id="team003",
                    name="Gamma Team",
                    organization="University C",
                    group=["graduate", "international"],
                    extra={"sponsor": "Tech Corp"},
                ),
            ]
        )

    def test_teams_basic_operations(self, sample_teams: Teams):
        """Test basic Teams operations"""
        assert len(sample_teams.root) == 3

        # Test iteration
        team_ids = [team.id for team in sample_teams.root]
        assert "team001" in team_ids
        assert "team002" in team_ids
        assert "team003" in team_ids

    def test_teams_serialization(self, sample_teams: Teams):
        """Test Teams serialization and deserialization"""
        # Test JSON serialization
        teams_json = sample_teams.model_dump_json()
        assert isinstance(teams_json, str)

        # Verify it's a valid JSON array
        parsed = json.loads(teams_json)
        assert isinstance(parsed, list)
        assert len(parsed) == 3

        # Test JSON deserialization
        reconstructed_teams = Teams.model_validate_json(teams_json)
        assert len(reconstructed_teams.root) == 3

        # Verify the data is correct
        reconstructed_teams.root.sort(key=lambda x: x.id)
        assert reconstructed_teams.root[0].id == "team001"
        assert reconstructed_teams.root[1].id == "team002"
        assert reconstructed_teams.root[2].id == "team003"

    def test_teams_round_trip(self, sample_teams: Teams):
        """Test complete round-trip serialization"""
        # Dict round-trip
        teams_dict = sample_teams.model_dump()
        reconstructed_from_dict = Teams.model_validate(teams_dict)
        assert len(reconstructed_from_dict.root) == len(sample_teams.root)

        # JSON round-trip
        teams_json = sample_teams.model_dump_json()
        reconstructed_from_json = Teams.model_validate_json(teams_json)
        assert len(reconstructed_from_json.root) == len(sample_teams.root)

    def test_empty_teams(self):
        """Test operations with empty Teams"""
        empty_teams = Teams([])

        assert len(empty_teams.root) == 0

        # Test serialization
        teams_json = empty_teams.model_dump_json()
        assert teams_json == "[]"

        # Test deserialization
        reconstructed = Teams.model_validate_json("[]")
        assert len(reconstructed.root) == 0



from xcpcio import constants
from xcpcio.types import BalloonColor, ContestOptions, Image, SubmissionReaction


class TestImage:
    """Test cases for Image Pydantic model"""

    def test_image_creation_empty(self):
        """Test Image creation with default values"""
        image = Image()
        assert image.url is None
        assert image.base64 is None
        assert image.preset is None

    def test_image_creation_with_values(self):
        """Test Image creation with provided values"""
        image = Image(url="https://example.com/image.png", base64="base64data", preset="ICPC")
        assert image.url == "https://example.com/image.png"
        assert image.base64 == "base64data"
        assert image.preset == "ICPC"

    def test_image_serialization(self):
        """Test Image serialization and deserialization"""
        image = Image(url="https://test.com/img.jpg", preset="CCPC")

        # Test model_dump
        image_dict = image.model_dump()
        assert image_dict["url"] == "https://test.com/img.jpg"
        assert image_dict["preset"] == "CCPC"
        assert image_dict["base64"] is None

        # Test JSON serialization
        image_json = image.model_dump_json()
        reconstructed_image = Image.model_validate_json(image_json)
        assert reconstructed_image == image


class TestBalloonColor:
    """Test cases for BalloonColor Pydantic model"""

    def test_color_creation(self):
        """Test BalloonColor creation"""
        color = BalloonColor(color="#ffffff", background_color="#000000")
        assert color.color == "#ffffff"
        assert color.background_color == "#000000"

    def test_color_serialization(self):
        """Test BalloonColor serialization and deserialization"""
        color = BalloonColor(color="red", background_color="blue")

        # Test model_dump
        color_dict = color.model_dump()
        assert color_dict["color"] == "red"
        assert color_dict["background_color"] == "blue"

        # Test JSON round-trip
        color_json = color.model_dump_json()
        reconstructed_color = BalloonColor.model_validate_json(color_json)
        assert reconstructed_color == color


class TestContestOptions:
    """Test cases for ContestOptions Pydantic model"""

    def test_contest_options_empty(self):
        """Test ContestOptions with default values"""
        options = ContestOptions()
        assert options.calculation_of_penalty is None
        assert options.submission_timestamp_unit is None
        assert options.has_reaction_videos is None
        assert options.reaction_video_url_template is None

    def test_contest_options_with_values(self):
        """Test ContestOptions with provided values"""
        options = ContestOptions(
            calculation_of_penalty=constants.CALCULATION_OF_PENALTY_IN_SECONDS,
            submission_timestamp_unit=constants.TIME_UNIT_MILLISECOND,
            has_reaction_videos=True,
            reaction_video_url_template="https://videos.com/{team_id}/{problem_id}",
        )

        assert options.calculation_of_penalty == constants.CALCULATION_OF_PENALTY_IN_SECONDS
        assert options.submission_timestamp_unit == constants.TIME_UNIT_MILLISECOND
        assert options.has_reaction_videos is True
        assert options.reaction_video_url_template == "https://videos.com/{team_id}/{problem_id}"

    def test_contest_options_serialization(self):
        """Test ContestOptions serialization"""
        options = ContestOptions(
            calculation_of_penalty=constants.CALCULATION_OF_PENALTY_IN_MINUTES, has_reaction_videos=False
        )

        # Test JSON round-trip
        options_json = options.model_dump_json()
        reconstructed_options = ContestOptions.model_validate_json(options_json)
        assert reconstructed_options == options


class TestSubmissionReaction:
    """Test cases for SubmissionReaction Pydantic model"""

    def test_reaction_empty(self):
        """Test SubmissionReaction with default values"""
        reaction = SubmissionReaction(url="")
        assert reaction.url == ""

    def test_reaction_with_url(self):
        """Test SubmissionReaction with URL"""
        reaction = SubmissionReaction(url="https://reaction.com/video.mp4")
        assert reaction.url == "https://reaction.com/video.mp4"

    def test_reaction_serialization(self):
        """Test SubmissionReaction serialization"""
        reaction = SubmissionReaction(url="https://test.com/reaction.mp4")

        # Test model_dump
        reaction_dict = reaction.model_dump()
        assert reaction_dict["url"] == "https://test.com/reaction.mp4"

        # Test JSON round-trip
        reaction_json = reaction.model_dump_json()
        reconstructed_reaction = SubmissionReaction.model_validate_json(reaction_json)
        assert reconstructed_reaction == reaction

    def test_reaction_no_url_serialization(self):
        """Test SubmissionReaction serialization with no URL"""
        reaction = SubmissionReaction(url="")

        reaction_dict = reaction.model_dump()
        assert reaction_dict["url"] == ""

        # Test JSON round-trip
        reaction_json = reaction.model_dump_json()
        reconstructed_reaction = SubmissionReaction.model_validate_json(reaction_json)
        assert reconstructed_reaction == reaction



import asyncio
import hashlib
import json
import logging
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Dict, Optional

from pydantic import BaseModel

from xcpcio.api.client import ApiClient
from xcpcio.api.models import FileData, UploadBoardDataResp
from xcpcio.clics.clics_api_client import APICredentials, ClicsApiClient, ClicsApiConfig

logger = logging.getLogger(__name__)


class FileChecksumEntry(BaseModel):
    checksum: str
    updated_at: str


class FileChecksumCacheData(BaseModel):
    checksums: Dict[str, FileChecksumEntry]
    updated_at: str


class FileChecksumCache:
    def __init__(
        self, cache_dir: Optional[Path] = None, contest_id: Optional[str] = None, version: Optional[str] = None
    ):
        if cache_dir:
            cache_dir = Path(cache_dir)
        else:
            cache_dir = Path.home() / ".xcpcio"

        cache_dir.mkdir(parents=True, exist_ok=True)

        if contest_id:
            self.cache_file = cache_dir / f"clics_uploader_{contest_id}.json"
        else:
            self.cache_file = cache_dir / "clics_uploader_cache.json"

        self.version_file = cache_dir / "version"
        self.version = version
        self.checksums: Dict[str, FileChecksumEntry] = {}
        self._load_cache()

    def _load_version(self) -> Optional[str]:
        if self.version_file.exists():
            try:
                with open(self.version_file, "r") as f:
                    return f.read().strip()
            except Exception as e:
                logger.warning(f"Failed to read version file: {e}")
        return None

    def _save_version(self):
        if not self.version:
            return
        try:
            with open(self.version_file, "w") as f:
                f.write(self.version)
        except Exception as e:
            logger.warning(f"Failed to save version file: {e}")

    def _load_cache(self):
        cached_version = self._load_version()
        if cached_version != self.version:
            logger.info(f"Cache version mismatch (cached: {cached_version}, current: {self.version}), clearing cache")
            self.checksums = {}
            return

        if self.cache_file.exists():
            try:
                with open(self.cache_file, "r") as f:
                    data = json.load(f)

                cache_data = FileChecksumCacheData.model_validate(data)
                self.checksums = cache_data.checksums
                logger.debug(f"Loaded cache with {len(self.checksums)} entries")
            except Exception as e:
                logger.warning(f"Failed to load cache: {e}")
                self.checksums = {}

    def _save_cache(self):
        try:
            self.cache_file.parent.mkdir(parents=True, exist_ok=True)
            cache_data = FileChecksumCacheData(
                checksums=self.checksums,
                updated_at=datetime.now().isoformat(),
            )
            with open(self.cache_file, "w") as f:
                json.dump(cache_data.model_dump(), f, indent=2)
            self._save_version()
            logger.debug(f"Saved cache with {len(self.checksums)} entries")
        except Exception as e:
            logger.warning(f"Failed to save cache: {e}")

    def calculate_checksum(self, content: str) -> str:
        return hashlib.sha256(content.encode("utf-8")).hexdigest()

    def has_changed(self, key: str, content: str) -> bool:
        checksum = self.calculate_checksum(content)
        old_entry = self.checksums.get(key)
        return not old_entry or old_entry.checksum != checksum

    def update_checksum(self, key: str, content: str):
        checksum = self.calculate_checksum(content)
        self.checksums[key] = FileChecksumEntry(
            checksum=checksum,
            updated_at=datetime.now().isoformat(),
        )

    def save(self):
        self._save_cache()


@dataclass
class UploaderConfig:
    clics_base_url: str
    contest_id: str
    clics_credentials: APICredentials
    xcpcio_api_url: str
    xcpcio_api_token: str
    cache_dir: Optional[Path] = None
    timeout: int = 30
    max_concurrent: int = 10
    poll_interval: int = 5
    version: Optional[str] = None

    def to_clics_api_config(self) -> ClicsApiConfig:
        return ClicsApiConfig(
            base_url=self.clics_base_url,
            credentials=self.clics_credentials,
            timeout=self.timeout,
            max_concurrent=self.max_concurrent,
        )


class ContestUploader:
    ENDPOINTS_TO_FETCH = (
        "access",
        "accounts",
        "awards",
        "groups",
        "judgement-types",
        "judgements",
        "languages",
        "organizations",
        "problems",
        "state",
        "submissions",
        "teams",
    )

    def __init__(self, config: UploaderConfig):
        self._config = config
        self._clics_client = ClicsApiClient(config.to_clics_api_config())
        self._api_client = ApiClient(
            base_url=config.xcpcio_api_url,
            token=config.xcpcio_api_token,
            timeout=config.timeout,
        )
        self._cache = FileChecksumCache(
            cache_dir=config.cache_dir,
            contest_id=config.contest_id,
            version=config.version,
        )

    async def __aenter__(self):
        await self._clics_client.__aenter__()
        await self._api_client.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._clics_client.__aexit__(exc_type, exc_val, exc_tb)
        await self._api_client.__aexit__(exc_type, exc_val, exc_tb)

    def _process_file_validations(
        self, file_map: Dict[str, FileData], file_keys: list[str], response: UploadBoardDataResp
    ):
        for file_key in file_keys:
            if file_key in file_map:
                validation = response.file_validations.get(file_key)
                if validation:
                    if validation.checksum_valid:
                        self._cache.update_checksum(file_key, file_map[file_key].content)
                    else:
                        logger.warning(f"File checksum validation failed for {file_key}: {validation.message}")
                else:
                    logger.warning(f"No validation result for {file_key}, will retry on next poll")

    async def _fetch_api_data(
        self,
        api_files: Dict[str, FileData],
        changed_api_files: list[str],
        unchanged_files: list[str],
    ):
        api_data = await self._clics_client.fetch_api_info()
        if api_data:
            filename = "api.json"
            content = json.dumps(api_data, ensure_ascii=False)

            if self._cache.has_changed(filename, content):
                checksum = self._cache.calculate_checksum(content)
                api_files[filename] = FileData(content=content, checksum=checksum)
                changed_api_files.append(filename)
            else:
                unchanged_files.append(filename)

    async def _fetch_contest_data(
        self,
    ) -> tuple[Dict[str, FileData], list[str], list[str]]:
        api_files = {}
        changed_api_files = []
        unchanged_files = []

        contest_data = await self._clics_client.fetch_json(f"contests/{self._config.contest_id}")
        if contest_data:
            filename = "contest.json"
            content = json.dumps(contest_data, ensure_ascii=False)

            if self._cache.has_changed(filename, content):
                checksum = self._cache.calculate_checksum(content)
                api_files[filename] = FileData(content=content, checksum=checksum)
                changed_api_files.append(filename)
            else:
                unchanged_files.append(filename)

        return api_files, changed_api_files, unchanged_files

    async def _fetch_endpoint_data(
        self,
        api_files: Dict[str, FileData],
        changed_api_files: list[str],
        unchanged_files: list[str],
    ):
        for endpoint in self.ENDPOINTS_TO_FETCH:
            data = await self._clics_client.fetch_json(f"contests/{self._config.contest_id}/{endpoint}")
            if data is not None:
                filename = f"{endpoint}.json"
                content = json.dumps(data, ensure_ascii=False)

                if self._cache.has_changed(filename, content):
                    checksum = self._cache.calculate_checksum(content)
                    api_files[filename] = FileData(content=content, checksum=checksum)
                    changed_api_files.append(filename)
                else:
                    unchanged_files.append(filename)

    async def _upload_api_files(
        self, api_files: Dict[str, FileData], changed_api_files: list[str]
    ) -> Optional[UploadBoardDataResp]:
        if not api_files:
            return None

        logger.info(f"Uploading {len(api_files)} API file(s) to XCPCIO...")
        response = await self._api_client.upload_board_data(extra_files=api_files)
        logger.info(f"API upload successful! ({len(api_files)} files)")

        self._process_file_validations(api_files, changed_api_files, response)
        self._cache.save()
        return response

    async def fetch_and_upload(self) -> Optional[Dict]:
        api_files, changed_api_files, unchanged_files = ({}, [], [])

        await self._fetch_api_data(api_files, changed_api_files, unchanged_files)

        contest_files, contest_changed, contest_unchanged = await self._fetch_contest_data()
        api_files.update(contest_files)
        changed_api_files.extend(contest_changed)
        unchanged_files.extend(contest_unchanged)

        await self._fetch_endpoint_data(api_files, changed_api_files, unchanged_files)

        if changed_api_files:
            logger.info(f"Changed API files ({len(changed_api_files)}): {', '.join(changed_api_files)}")
        if unchanged_files:
            logger.debug(f"Unchanged files ({len(unchanged_files)}): {', '.join(unchanged_files)}")

        if not api_files:
            logger.info("No changed files to upload")
            return None

        response = await self._upload_api_files(api_files, changed_api_files)

        logger.info(f"All uploads completed! ({len(api_files)} files)")
        return response.model_dump() if response else None

    async def run_loop(self):
        iteration = 0
        while True:
            iteration += 1
            try:
                logger.info(f"[Iteration #{iteration}] Starting fetch and upload...")
                await self.fetch_and_upload()
            except Exception as e:
                logger.exception(f"Error during fetch and upload: {e}", exc_info=True)

            logger.info(f"Waiting {self._config.poll_interval} seconds until next poll...")
            await asyncio.sleep(self._config.poll_interval)



import logging
from typing import Any, Dict, List

from fastapi import APIRouter, Path

from ..dependencies import ContestServiceDep

router = APIRouter()
logger = logging.getLogger(__name__)


@router.get(
    "/contests/{contest_id}/judgement-types",
    summary="Get all the judgement types for this contest",
    response_model=List[Dict[str, Any]],
)
async def get_judgement_types(
    contest_id: str = Path(..., description="Contest identifier"),
    service: ContestServiceDep = None,
) -> List[Dict[str, Any]]:
    return service.get_judgement_types(contest_id)


@router.get(
    "/contests/{contest_id}/judgement-types/{judgement_type_id}",
    summary="Get the given judgement type for this contest",
    response_model=Dict[str, Any],
)
async def get_judgement_type(
    contest_id: str = Path(..., description="Contest identifier"),
    judgement_type_id: str = Path(..., description="Judgement type identifier"),
    service: ContestServiceDep = None,
) -> Dict[str, Any]:
    return service.get_judgement_type(contest_id, judgement_type_id)
